"""
Animation module for creating complex, composable animations.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from .compose import AnimationComposerDialog

__all__ = ["AnimationComposerDialog"]
